evview {
    background-color: @theme_bg_color;
}

evview.document-page {
	background-color: @theme_bg_color;
    border-style: solid;
    border-width: 3px 3px 6px 4px;
    border-image: url("resource:///org/x/reader/previewer/ui/thumbnail-frame.png") 3 3 6 4;
}

evview.document-page.inverted {
    background-color: black;
}
